"""
Backport of re.escape() from Python 3.7.

The behavior changed in Python 3.7 so we need this to get a consistent
behaviour until we drop support for Python 3.6.

> Change re.escape() to only escape regex special characters instead of
> escaping all characters other than ASCII letters, numbers, and '_'.

Source: https://docs.python.org/3/whatsnew/3.7.html
"""

_special_chars_map = {i: "\\" + chr(i) for i in b"()[]{}?*+-|^$\\.&~# \t\n\r\v\f"}


def escape(pattern):
    """Escape special characters in a string."""
    return pattern.translate(_special_chars_map)



from typing import Optional, Union

from libcst import (
    Arg,
    Attribute,
    BaseExpression,
    BaseSmallStatement,
    BaseStatement,
    Call,
    FunctionDef,
    ImportFrom,
    MaybeSentinel,
    Name,
    RemovalSentinel,
    RemoveFromParent,
    Return,
)
from libcst import matchers as m
from libcst.codemod.visitors import AddImportsVisitor

from django_codemod.constants import DJANGO_1_9, DJANGO_1_11, DJANGO_2_0, DJANGO_2_1
from django_codemod.visitors.base import BaseDjCodemodTransformer, module_matcher


class ModelsPermalinkTransformer(BaseDjCodemodTransformer):
    """Replace `@models.permalink` decorator by a call to `reverse()`."""

    deprecated_in = DJANGO_1_11
    removed_in = DJANGO_2_1
    ctx_key_prefix = "ModelsPermalinkTransformer"
    ctx_key_inside_method = f"{ctx_key_prefix}-inside_method"
    ctx_key_decorator_matchers = f"{ctx_key_prefix}-decorator_matchers"

    def leave_ImportFrom(
        self, original_node: ImportFrom, updated_node: ImportFrom
    ) -> Union[BaseSmallStatement, RemovalSentinel]:
        if m.matches(
            updated_node,
            m.ImportFrom(module=module_matcher(["django", "db"])),
        ):
            for imported_name in updated_node.names:
                if m.matches(imported_name, m.ImportAlias(name=m.Name("models"))):
                    self.add_decorator_matcher(
                        m.Decorator(
                            decorator=m.Attribute(
                                value=m.Name("models"), attr=m.Name("permalink")
                            )
                        )
                    )
        if m.matches(
            updated_node,
            m.ImportFrom(module=module_matcher(["django", "db", "models"])),
        ):
            updated_names = []
            for imported_name in updated_node.names:
                if m.matches(imported_name, m.ImportAlias(name=m.Name("permalink"))):
                    decorator_name = (
                        imported_name.asname.name.value
                        if imported_name.asname
                        else "permalink"
                    )
                    self.add_decorator_matcher(
                        m.Decorator(decorator=m.Name(decorator_name))
                    )
                else:
                    updated_names.append(imported_name)
            if not updated_names:
                return RemoveFromParent()
            # sort imports
            new_names = sorted(updated_names, key=lambda n: n.evaluated_name)
            # remove any trailing commas
            last_name = new_names[-1]
            if last_name.comma != MaybeSentinel.DEFAULT:
                new_names[-1] = last_name.with_changes(comma=MaybeSentinel.DEFAULT)
            return updated_node.with_changes(names=new_names)
        return super().leave_ImportFrom(original_node, updated_node)

    def add_decorator_matcher(self, matcher):
        if self.ctx_key_decorator_matchers not in self.context.scratch:
            self.context.scratch[self.ctx_key_decorator_matchers] = []
        self.context.scratch[self.ctx_key_decorator_matchers].append(matcher)

    @property
    def decorator_matcher(self):
        matchers_list = self.context.scratch.get(self.ctx_key_decorator_matchers, [])
        if len(matchers_list) == 0:
            return None
        if len(matchers_list) == 1:
            return matchers_list[0]
        return m.OneOf(*[matcher for matcher in matchers_list])

    def visit_FunctionDef(self, node: FunctionDef) -> Optional[bool]:
        for decorator in node.decorators:
            if m.matches(decorator, self.decorator_matcher):
                self.context.scratch[self.ctx_key_inside_method] = True
        return super().visit_FunctionDef(node)

    def leave_FunctionDef(
        self, original_node: FunctionDef, updated_node: FunctionDef
    ) -> Union[BaseStatement, RemovalSentinel]:
        if self.visiting_permalink_method:
            for decorator in updated_node.decorators:
                if m.matches(decorator, self.decorator_matcher):
                    AddImportsVisitor.add_needed_import(
                        context=self.context,
                        module="django.urls",
                        obj="reverse",
                    )
                    updated_decorators = list(updated_node.decorators)
                    updated_decorators.remove(decorator)
                    self.context.scratch.pop(self.ctx_key_inside_method, None)
                    return updated_node.with_changes(
                        decorators=tuple(updated_decorators)
                    )
        return super().leave_FunctionDef(original_node, updated_node)

    @property
    def visiting_permalink_method(self):
        return self.context.scratch.get(self.ctx_key_inside_method, False)

    def leave_Return(
        self, original_node: Return, updated_node: Return
    ) -> Union[BaseSmallStatement, RemovalSentinel]:
        if self.visiting_permalink_method and m.matches(updated_node.value, m.Tuple()):
            elem_0 = updated_node.value.elements[0]
            elem_1_3 = updated_node.value.elements[1:3]
            args = (
                Arg(elem_0.value),
                Arg(Name("None")),
                *[Arg(el.value) for el in elem_1_3],
            )
            return updated_node.with_changes(
                value=Call(func=Name("reverse"), args=args)
            )
        return super().leave_Return(original_node, updated_node)


def is_foreign_key(node: Call) -> bool:
    return m.matches(node, m.Call(func=m.Attribute(attr=m.Name(value="ForeignKey"))))


def is_one_to_one_field(node: Call) -> bool:
    return m.matches(
        node,
        m.Call(func=m.Attribute(attr=m.Name(value="OneToOneField"))),
    )


def has_on_delete(node: Call) -> bool:
    # if on_delete exists in any kwarg we return True
    for arg in node.args:
        if m.matches(arg, m.Arg(keyword=m.Name("on_delete"))):
            return True

    # if there are two or more nodes and there are no keywords
    # then we can assume that positional arguments are being used
    # and on_delete is being handled.
    return len(node.args) >= 2 and node.args[1].keyword is None


class OnDeleteTransformer(BaseDjCodemodTransformer):
    """Add the `on_delete=CASCADE` to `ForeignKey` and `OneToOneField`."""

    deprecated_in = DJANGO_1_9
    removed_in = DJANGO_2_0
    ctx_key_prefix = "OnDeleteTransformer"

    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        if (
            is_one_to_one_field(original_node) or is_foreign_key(original_node)
        ) and not has_on_delete(original_node):
            AddImportsVisitor.add_needed_import(
                context=self.context,
                module="django.db",
                obj="models",
            )
            updated_args = (
                *updated_node.args,
                Arg(
                    keyword=Name("on_delete"),
                    value=Attribute(value=Name("models"), attr=Name("CASCADE")),
                ),
            )
            return updated_node.with_changes(args=updated_args)
        return super().leave_Call(original_node, updated_node)



from django_codemod.constants import DJANGO_1_10, DJANGO_2_0
from django_codemod.visitors.base import BaseModuleRenameTransformer


class URLResolversTransformer(BaseModuleRenameTransformer):
    """Replace `django.core.urlresolvers` by `django.urls`."""

    deprecated_in = DJANGO_1_10
    removed_in = DJANGO_2_0
    rename_from = "django.core.urlresolvers"
    rename_to = "django.urls"



from typing import Sequence

from libcst import Arg, Call, Name

from django_codemod.constants import DJANGO_2_0, DJANGO_3_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class RenderToResponseTransformer(BaseFuncRenameTransformer):
    """
    Replace `render_to_response()` by `render()`.

    Use `None` as the first argument to `render()`.
    """

    deprecated_in = DJANGO_2_0
    removed_in = DJANGO_3_0
    rename_from = "django.shortcuts.render_to_response"
    rename_to = "django.shortcuts.render"

    def update_call_args(self, node: Call) -> Sequence[Arg]:
        return (Arg(value=Name("None")), *node.args)



from libcst.codemod import CodemodTest

from django_codemod.commands import BaseCodemodCommand


class BaseVisitorTest(CodemodTest):
    """Base test to use in all visitors tests."""

    transformer = None

    def TRANSFORM(self, context):
        """Create a command for the transformer under test."""
        return BaseCodemodCommand([self.transformer], context)



from django_codemod.visitors import LRUCacheTransformer
from tests.visitors.base import BaseVisitorTest


class TestLRUCacheTransformer(BaseVisitorTest):

    transformer = LRUCacheTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.lru_cache import lru_cache

            result = lru_cache(content)
        """
        after = """
            from functools import lru_cache

            result = lru_cache(content)
        """
        self.assertCodemod(before, after)



from typing import Union

from libcst import (
    Arg,
    BaseExpression,
    BaseSmallStatement,
    BaseStatement,
    Call,
    ClassDef,
    FunctionDef,
    ImportFrom,
    Module,
    Name,
    Param,
    RemovalSentinel,
)
from libcst import matchers as m

from django_codemod.constants import DJANGO_2_1, DJANGO_3_0
from django_codemod.visitors.base import BaseDjCodemodTransformer, module_matcher


class InlineHasAddPermissionsTransformer(BaseDjCodemodTransformer):
    """Add the `obj` argument to `InlineModelAdmin.has_add_permission()`."""

    deprecated_in = DJANGO_2_1
    removed_in = DJANGO_3_0
    ctx_key_prefix = "InlineHasAddPermissionsTransformer"
    ctx_key_base_cls_matcher = f"{ctx_key_prefix}-base_cls_matcher"
    ctx_key_visiting_subclass = f"{ctx_key_prefix}-is_visiting_subclass"

    def leave_ImportFrom(
        self, original_node: ImportFrom, updated_node: ImportFrom
    ) -> Union[BaseSmallStatement, RemovalSentinel]:
        base_cls_matcher = []
        if m.matches(
            updated_node,
            m.ImportFrom(module=module_matcher(["django", "contrib", "admin"])),
        ):
            for imported_name in updated_node.names:
                if m.matches(
                    imported_name, m.ImportAlias(name=m.Name("TabularInline"))
                ):
                    base_cls_matcher.append(m.Arg(m.Name("TabularInline")))
                if m.matches(
                    imported_name, m.ImportAlias(name=m.Name("StackedInline"))
                ):
                    base_cls_matcher.append(m.Arg(m.Name("StackedInline")))
        if m.matches(
            updated_node,
            m.ImportFrom(module=module_matcher(["django", "contrib"])),
        ):
            for imported_name in updated_node.names:
                if m.matches(imported_name, m.ImportAlias(name=m.Name("admin"))):

                    base_cls_matcher.extend(
                        [
                            m.Arg(
                                m.Attribute(
                                    value=m.Name("admin"), attr=m.Name("TabularInline")
                                )
                            ),
                            m.Arg(
                                m.Attribute(
                                    value=m.Name("admin"), attr=m.Name("StackedInline")
                                )
                            ),
                        ]
                    )
        # Save valid matchers in the context
        if base_cls_matcher:
            self.context.scratch[self.ctx_key_base_cls_matcher] = m.OneOf(
                *base_cls_matcher
            )
        return super().leave_ImportFrom(original_node, updated_node)

    def leave_Module(self, original_node: Module, updated_node: Module) -> Module:
        self.context.scratch.pop(self.ctx_key_base_cls_matcher, None)
        return super().leave_Module(original_node, updated_node)

    @property
    def base_cls_matcher(self):
        return self.context.scratch.get(self.ctx_key_base_cls_matcher)

    def visit_ClassDef_bases(self, node: ClassDef) -> None:
        if self.base_cls_matcher is not None:
            for base_cls in node.bases:
                if m.matches(base_cls, self.base_cls_matcher):
                    self.context.scratch[self.ctx_key_visiting_subclass] = True
        super().visit_ClassDef_bases(node)

    def leave_ClassDef(
        self, original_node: ClassDef, updated_node: ClassDef
    ) -> Union[BaseStatement, RemovalSentinel]:
        self.context.scratch.pop(self.ctx_key_visiting_subclass, None)
        return super().leave_ClassDef(original_node, updated_node)

    @property
    def is_visiting_subclass(self):
        return self.context.scratch.get(self.ctx_key_visiting_subclass, False)

    def leave_FunctionDef(
        self, original_node: FunctionDef, updated_node: FunctionDef
    ) -> Union[BaseStatement, RemovalSentinel]:
        if (
            self.is_visiting_subclass
            and m.matches(
                updated_node, m.FunctionDef(name=m.Name("has_add_permission"))
            )
            and len(updated_node.params.params) == 2
        ):
            old_params = updated_node.params
            updated_params = old_params.with_changes(
                params=(
                    *old_params.params,
                    Param(name=Name("obj"), default=Name("None")),
                )
            )
            return updated_node.with_changes(params=updated_params)
        return super().leave_FunctionDef(original_node, updated_node)

    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        if (
            self.is_visiting_subclass
            and m.matches(
                updated_node,
                m.Call(
                    func=m.Attribute(
                        attr=m.Name("has_add_permission"),
                        value=m.Call(func=m.Name("super")),
                    )
                ),
            )
            and len(updated_node.args) < 2
        ):
            updated_args = (
                *updated_node.args,
                Arg(keyword=Name("obj"), value=Name("obj")),
            )
            return updated_node.with_changes(args=updated_args)
        return super().leave_Call(original_node, updated_node)



from django_codemod.visitors.decorators import (
    AvailableAttrsTransformer,
    ContextDecoratorTransformer,
)
from tests.visitors.base import BaseVisitorTest


class TestContextDecoratorTransformer(BaseVisitorTest):

    transformer = ContextDecoratorTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.decorators import ContextDecorator

            class mycontext(ContextDecorator):
                def __enter__(self):
                    print('Starting')
                    return self

                def __exit__(self, *exc):
                    print('Finishing')
                    return False
        """
        after = """
            from contextlib import ContextDecorator

            class mycontext(ContextDecorator):
                def __enter__(self):
                    print('Starting')
                    return self

                def __exit__(self, *exc):
                    print('Finishing')
                    return False
        """
        self.assertCodemod(before, after)


class TestAvailableAttrsTransformer(BaseVisitorTest):

    transformer = AvailableAttrsTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.decorators import available_attrs

            def my_decorator(func):
                @wraps(func, assigned=available_attrs(func))
                def inner(*args, **kwargs):
                    return func(*args, **kwargs)

                return inner
        """
        after = """
            from functools import WRAPPER_ASSIGNMENTS

            def my_decorator(func):
                @wraps(func, assigned=WRAPPER_ASSIGNMENTS)
                def inner(*args, **kwargs):
                    return func(*args, **kwargs)

                return inner
        """
        self.assertCodemod(before, after)



"""Module to implement base functionality."""
from abc import ABC
from typing import Generator, Optional, Sequence, Tuple, Union

from libcst import (
    Arg,
    BaseExpression,
    Call,
    ImportStar,
    MaybeSentinel,
    Name,
    RemovalSentinel,
    RemoveFromParent,
)
from libcst import matchers as m
from libcst._nodes.statement import BaseSmallStatement, ImportAlias, ImportFrom
from libcst.codemod import ContextAwareTransformer
from libcst.codemod.visitors import AddImportsVisitor


class BaseDjCodemodTransformer(ContextAwareTransformer, ABC):
    deprecated_in: Tuple[int, int]
    removed_in: Tuple[int, int]


def module_matcher(import_parts):
    *values, attr = import_parts
    if len(values) > 1:
        value = module_matcher(values)
    elif len(values) == 1:
        value = m.Name(values[0])
    else:
        value = m.DoNotCare()
    return m.Attribute(value=value, attr=m.Name(attr))


def import_from_matches(node, module_parts):
    return m.matches(node, m.ImportFrom(module=module_matcher(module_parts)))


class BaseRenameTransformer(BaseDjCodemodTransformer, ABC):
    """Base class to help rename or move a declaration."""

    rename_from: str
    rename_to: str

    simple_rename = True

    @property
    def old_name(self):
        return self.rename_from.split(".")[-1]

    @property
    def old_module_parts(self):
        return self.rename_from.split(".")[:-1]

    @property
    def new_name(self):
        return self.rename_to.split(".")[-1]

    @property
    def new_module_parts(self):
        return self.rename_to.split(".")[:-1]

    @property
    def ctx_key_imported_as(self):
        return f"{self.rename_from}-imported_as"

    @property
    def entity_imported_as(self):
        return self.context.scratch.get(self.ctx_key_imported_as, None)

    @property
    def is_imported_with_old_name(self):
        is_imported = self.ctx_key_imported_as in self.context.scratch
        return is_imported and not self.entity_imported_as

    def leave_ImportFrom(
        self, original_node: ImportFrom, updated_node: ImportFrom
    ) -> Union[BaseSmallStatement, RemovalSentinel]:
        """Update import statements for matching old module name."""
        if not import_from_matches(updated_node, self.old_module_parts):
            return super().leave_ImportFrom(original_node, updated_node)
        # This is a match
        new_names = list(self.gen_new_imported_names(updated_node.names))
        if not new_names:
            # Nothing left in the import statement: remove it
            return RemoveFromParent()
        # Some imports are left, update the statement
        cleaned_names = self.tidy_new_imported_names(new_names)
        return updated_node.with_changes(names=cleaned_names)

    def gen_new_imported_names(
        self, old_names: Union[Sequence[ImportAlias], ImportStar]
    ) -> Generator:
        """Update import if the entity we're interested in is imported."""
        for import_alias in old_names:
            if not self.old_name or import_alias.evaluated_name == self.old_name:
                self.context.scratch[self.ctx_key_imported_as] = import_alias.asname
                if self.simple_rename:
                    self.add_new_import(import_alias.evaluated_name)
            else:
                yield import_alias

    def tidy_new_imported_names(self, new_names):
        """Tidy up the updated list of imports"""
        # Sort them
        cleaned_names = sorted(new_names, key=lambda n: n.evaluated_name)
        # Remove any trailing commas
        last_name = cleaned_names[-1]
        if last_name.comma != MaybeSentinel.DEFAULT:
            cleaned_names[-1] = last_name.with_changes(comma=MaybeSentinel.DEFAULT)
        return cleaned_names

    def add_new_import(self, evaluated_name: Optional[str] = None):
        as_name = (
            self.entity_imported_as.name.value if self.entity_imported_as else None
        )
        AddImportsVisitor.add_needed_import(
            context=self.context,
            module=".".join(self.new_module_parts),
            obj=self.new_name or evaluated_name,
            asname=as_name,
        )


class BaseModuleRenameTransformer(BaseRenameTransformer, ABC):
    """Base class to help rename or move a module."""

    @property
    def old_name(self):
        return ""

    @property
    def old_module_parts(self):
        return self.rename_from.split(".")

    @property
    def new_name(self):
        return ""

    @property
    def new_module_parts(self):
        return self.rename_to.split(".")


class BaseFuncRenameTransformer(BaseRenameTransformer, ABC):
    """Base class to help rename or move a function."""

    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        if self.is_imported_with_old_name and m.matches(
            updated_node, m.Call(func=m.Name(self.old_name))
        ):
            return self.update_call(updated_node=updated_node)
        return super().leave_Call(original_node, updated_node)

    def update_call(self, updated_node: Call) -> BaseExpression:
        updated_args = self.update_call_args(updated_node)
        return Call(args=updated_args, func=Name(self.new_name))

    def update_call_args(self, node: Call) -> Sequence[Arg]:
        return node.args



from django_codemod.visitors import (
    HttpUrlQuotePlusTransformer,
    HttpUrlQuoteTransformer,
    HttpUrlUnQuotePlusTransformer,
    HttpUrlUnQuoteTransformer,
    IsSafeUrlTransformer,
)
from tests.visitors.base import BaseVisitorTest


class TestHttpUrlQuoteTransformer(BaseVisitorTest):

    transformer = HttpUrlQuoteTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.http import urlquote

            result = urlquote(content)
        """
        after = """
            from urllib.parse import quote

            result = quote(content)
        """
        self.assertCodemod(before, after)


class TestHttpUrlQuotePlusTransformer(BaseVisitorTest):

    transformer = HttpUrlQuotePlusTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.http import urlquote_plus

            result = urlquote_plus(content)
        """
        after = """
            from urllib.parse import quote_plus

            result = quote_plus(content)
        """
        self.assertCodemod(before, after)


class TestHttpUrlUnQuoteTransformer(BaseVisitorTest):

    transformer = HttpUrlUnQuoteTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.http import urlunquote

            result = urlunquote(content)
        """
        after = """
            from urllib.parse import unquote

            result = unquote(content)
        """
        self.assertCodemod(before, after)


class TestHttpUrlUnQuotePlusTransformer(BaseVisitorTest):

    transformer = HttpUrlUnQuotePlusTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.http import urlunquote_plus

            result = urlunquote_plus(content)
        """
        after = """
            from urllib.parse import unquote_plus

            result = unquote_plus(content)
        """
        self.assertCodemod(before, after)


class TestIsSafeUrlTransformer(BaseVisitorTest):

    transformer = IsSafeUrlTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.http import is_safe_url

            result = is_safe_url('http://test.com/some-path', None)
        """
        after = """
            from django.utils.http import url_has_allowed_host_and_scheme

            result = url_has_allowed_host_and_scheme('http://test.com/some-path', None)
        """
        self.assertCodemod(before, after)



from django_codemod.visitors import RenderToResponseTransformer
from tests.visitors.base import BaseVisitorTest


class TestRenderToResponseTransformer(BaseVisitorTest):

    transformer = RenderToResponseTransformer

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.shortcuts import render_to_response

            result = render_to_response("index.html", context={}, status=None)
        """
        after = """
            from django.shortcuts import render

            result = render(None, "index.html", context={}, status=None)
        """
        self.assertCodemod(before, after)



from parameterized import parameterized

from django_codemod.visitors.models import (
    ModelsPermalinkTransformer,
    OnDeleteTransformer,
)
from tests.visitors.base import BaseVisitorTest


class TestOnDeleteTransformer(BaseVisitorTest):

    transformer = OnDeleteTransformer

    def test_foreign_key(self) -> None:
        before = """
            class MyModel(models.Model):
                user = models.ForeignKey('auth.User')
        """
        after = """
            from django.db import models

            class MyModel(models.Model):
                user = models.ForeignKey('auth.User', on_delete = models.CASCADE)
        """
        self.assertCodemod(before, after)

    def test_one_to_one_field(self) -> None:
        before = """
            class MyModel(models.Model):
                user = models.OneToOneField('auth.User')
        """
        after = """
            from django.db import models

            class MyModel(models.Model):
                user = models.OneToOneField('auth.User', on_delete = models.CASCADE)
        """
        self.assertCodemod(before, after)

    def test_multiple_kwargs_with_on_delete(self) -> None:
        """ Should do nothing """
        before = """
            from django.db import models

            class MyModel(models.Model):
                operations = [
                    migrations.CreateModel(
                        fields=[
                            (
                                'model',
                                models.OneToOneField(
                                    to = 'random.Model', on_delete = models.CASCADE),
                            ),
                        ],
                    )
                ]
        """
        after = """
            from django.db import models

            class MyModel(models.Model):
                operations = [
                    migrations.CreateModel(
                        fields=[
                            (
                                'model',
                                models.OneToOneField(
                                    to = 'random.Model', on_delete = models.CASCADE),
                            ),
                        ],
                    )
                ]
        """
        self.assertCodemod(before, after)

    def test_multiple_kwargs_without_on_delete(self) -> None:
        before = """
            from django.db import models


            class MyModel(models.Model):
                operations = [
                    migrations.AddField(
                        field=models.ForeignKey(
                            related_name='model',
                            blank=True,
                            to='random.Model',
                            null=True,
                        ),
                    )
                ]
        """
        after = """
            from django.db import models


            class MyModel(models.Model):
                operations = [
                    migrations.AddField(
                        field=models.ForeignKey(
                            related_name='model',
                            blank=True,
                            to='random.Model',
                            null=True,
                        on_delete = models.CASCADE),
                    )
                ]
        """
        self.assertCodemod(before, after)

    def test_positional_args_without_on_delete(self) -> None:
        before = """
            from django.db import models

            models.OneToOneField('model')
        """
        after = """
            from django.db import models

            models.OneToOneField('model', on_delete = models.CASCADE)
        """
        self.assertCodemod(before, after)

    def test_positional_args_with_on_delete(self) -> None:
        """ Should do nothing """
        before = """
            from django.db import models


            class MyModel(models.randomModel):
                operations = [
                    migrations.CreateModel(
                        fields=[
                            (
                                'field_name',
                                models.OneToOneField(
                                    'model.Location',
                                    models.CASCADE,
                                    auto_created=True,
                                ),
                            ),
                        ],
                    )
                ]
        """
        after = """
            from django.db import models


            class MyModel(models.randomModel):
                operations = [
                    migrations.CreateModel(
                        fields=[
                            (
                                'field_name',
                                models.OneToOneField(
                                    'model.Location',
                                    models.CASCADE,
                                    auto_created=True,
                                ),
                            ),
                        ],
                    )
                ]
        """
        self.assertCodemod(before, after)

    def test_add_model_import(self) -> None:
        before = """
            from random.place import CustomModel

            class MyModel(CustomModel):
                operations = [
                    migrations.CreateModel(
                        fields=[
                            (
                                'field_name',
                                CustomModel.OneToOneField(
                                    'model.Location',
                                    auto_created=True,
                                ),
                            ),
                        ],
                    )
                ]
        """
        after = """
            from random.place import CustomModel
            from django.db import models

            class MyModel(CustomModel):
                operations = [
                    migrations.CreateModel(
                        fields=[
                            (
                                'field_name',
                                CustomModel.OneToOneField(
                                    'model.Location',
                                    auto_created=True,
                                on_delete = models.CASCADE),
                            ),
                        ],
                    )
                ]
        """
        self.assertCodemod(before, after)


class TestAvailableAttrsTransformer(BaseVisitorTest):

    transformer = ModelsPermalinkTransformer

    def test_simple_substitution_args(self) -> None:
        before = """
            from django.db import models

            class MyModel(models.Model):

                @models.permalink
                def url(self):
                    return ('guitarist_detail', [self.slug])

                def get_name(self):
                    return get_name_from(self)
        """
        after = """
            from django.db import models
            from django.urls import reverse

            class MyModel(models.Model):

                def url(self):
                    return reverse('guitarist_detail', None, [self.slug])

                def get_name(self):
                    return get_name_from(self)
        """
        self.assertCodemod(before, after)

    def test_simple_substitution_kwargs(self) -> None:
        before = """
            from django.db import models

            class MyModel(models.Model):

                @models.permalink
                def url(self):
                    return ('guitarist_detail', [], {'slug': self.slug})
        """
        after = """
            from django.db import models
            from django.urls import reverse

            class MyModel(models.Model):

                def url(self):
                    return reverse('guitarist_detail', None, [], {'slug': self.slug})
        """
        self.assertCodemod(before, after)

    @parameterized.expand(
        [
            ("permalink", "permalink"),
            ("permalink as models_permalink", "models_permalink"),
        ]
    )
    def test_specific_import_alone(self, import_as, decorator_name) -> None:
        before = f"""
            from django.db import models
            from django.db.models import {import_as}

            class MyModel(models.Model):

                @{decorator_name}
                def url(self):
                    return ('guitarist_detail', [self.slug])
        """
        after = """
            from django.db import models
            from django.urls import reverse

            class MyModel(models.Model):

                def url(self):
                    return reverse('guitarist_detail', None, [self.slug])
        """
        self.assertCodemod(before, after)

    @parameterized.expand(
        [
            ("permalink, ObjectDoesNotExist", "permalink"),
            ("ObjectDoesNotExist, permalink", "permalink"),
            ("permalink as models_permalink, ObjectDoesNotExist", "models_permalink"),
            ("ObjectDoesNotExist, permalink as models_permalink", "models_permalink"),
        ]
    )
    def test_specific_import_with_others(self, import_as, decorator_name) -> None:
        before = f"""
            from django.db import models
            from django.db.models import {import_as}

            class MyModel(models.Model):

                @{decorator_name}
                def url(self):
                    return ('guitarist_detail', [self.slug])

                def get_name(self):
                    return 'World'
        """
        after = """
            from django.db import models
            from django.db.models import ObjectDoesNotExist
            from django.urls import reverse

            class MyModel(models.Model):

                def url(self):
                    return reverse('guitarist_detail', None, [self.slug])

                def get_name(self):
                    return 'World'
        """
        self.assertCodemod(before, after)



from django_codemod.visitors import URLResolversTransformer
from tests.visitors.base import BaseVisitorTest


class TestURLResolversTransformer(BaseVisitorTest):

    transformer = URLResolversTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.core.urlresolvers import reverse

            result = reverse('home_page')
        """
        after = """
            from django.urls import reverse

            result = reverse('home_page')
        """
        self.assertCodemod(before, after)



from django_codemod.constants import DJANGO_2_0, DJANGO_3_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class LRUCacheTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.lru_cache.lru_cache` by `functools.lru_cache`."""

    deprecated_in = DJANGO_2_0
    removed_in = DJANGO_3_0
    rename_from = "django.utils.lru_cache.lru_cache"
    rename_to = "functools.lru_cache"



from django_codemod.visitors import AbsPathTransformer
from tests.visitors.base import BaseVisitorTest


class TestAbsPathTransformer(BaseVisitorTest):

    transformer = AbsPathTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils._os import abspathu

            result = abspathu(content)
        """
        after = """
            from os.path import abspath

            result = abspath(content)
        """
        self.assertCodemod(before, after)



from .admin import InlineHasAddPermissionsTransformer
from .core import URLResolversTransformer
from .decorators import AvailableAttrsTransformer, ContextDecoratorTransformer
from .encoding import (
    ForceTextTransformer,
    SmartTextTransformer,
    UnicodeCompatibleTransformer,
)
from .html import UnescapeEntitiesTransformer
from .http import (
    HttpUrlQuotePlusTransformer,
    HttpUrlQuoteTransformer,
    HttpUrlUnQuotePlusTransformer,
    HttpUrlUnQuoteTransformer,
    IsSafeUrlTransformer,
)
from .lru_cache import LRUCacheTransformer
from .models import ModelsPermalinkTransformer, OnDeleteTransformer
from .os_utils import AbsPathTransformer
from .shortcuts import RenderToResponseTransformer
from .translations import (
    UGetTextLazyTransformer,
    UGetTextNoopTransformer,
    UGetTextTransformer,
    UNGetTextLazyTransformer,
    UNGetTextTransformer,
)
from .urls import URLTransformer

__all__ = (
    "AbsPathTransformer",
    "AvailableAttrsTransformer",
    "ContextDecoratorTransformer",
    "ForceTextTransformer",
    "HttpUrlQuotePlusTransformer",
    "HttpUrlQuoteTransformer",
    "HttpUrlUnQuotePlusTransformer",
    "HttpUrlUnQuoteTransformer",
    "InlineHasAddPermissionsTransformer",
    "IsSafeUrlTransformer",
    "LRUCacheTransformer",
    "ModelsPermalinkTransformer",
    "OnDeleteTransformer",
    "RenderToResponseTransformer",
    "SmartTextTransformer",
    "UGetTextLazyTransformer",
    "UGetTextNoopTransformer",
    "UGetTextTransformer",
    "UNGetTextLazyTransformer",
    "UNGetTextTransformer",
    "URLTransformer",
    "URLResolversTransformer",
    "UnescapeEntitiesTransformer",
    "UnicodeCompatibleTransformer",
)



from libcst import Arg, BaseExpression, Call, Name, SimpleString
from libcst import matchers as m
from libcst.codemod.visitors import AddImportsVisitor

from django_codemod import re
from django_codemod.constants import DJANGO_3_0, DJANGO_4_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class PatternNotSupported(RuntimeError):
    pass


REGEX_TO_CONVERTER = {
    "[0-9]+": "int",
    r"\d+": "int",
    ".+": "path",
    "[-a-zA-Z0-9_]+": "slug",
    r"[\w-]+": "slug",
    "[^/]+": "str",
    "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}": "uuid",
}


class URLTransformer(BaseFuncRenameTransformer):
    """Update `django.conf.urls.url` by equivalent `path` or `re_path`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.conf.urls.url"
    rename_to = "django.urls.re_path"

    simple_rename = False

    def update_call(self, updated_node: Call) -> BaseExpression:
        """Update `url` call with either `path` or `re_path`."""
        try:
            return self.update_call_to_path(updated_node)
        except PatternNotSupported:
            # Safe fallback to re_path()
            self.add_new_import()
            return super().update_call(updated_node)

    def update_call_to_path(self, updated_node: Call):
        """Update an URL pattern to `path()` in simple cases."""
        first_arg, *other_args = updated_node.args
        self.check_not_simple_string(first_arg)
        # Extract the URL pattern from the first argument
        pattern = first_arg.value.evaluated_value
        self.check_missing_start(pattern)
        # If we reach this point, we might be able to use `path()`
        call = self.build_path_call(pattern, other_args)
        AddImportsVisitor.add_needed_import(
            context=self.context,
            module=".".join(self.new_module_parts),
            obj="path",
        )
        return call

    def check_not_simple_string(self, first_arg: Arg):
        """Translated patterns are not supported."""
        if not m.matches(first_arg, m.Arg(value=m.SimpleString())):
            raise PatternNotSupported()

    def check_missing_start(self, pattern):
        """Patterns that do not match the start of the string with caret."""
        if not pattern.startswith("^"):
            raise PatternNotSupported()

    def build_path_call(self, pattern, other_args):
        """Build the `Call` node using Django 2.0's `path()` function."""
        route = self.build_route(pattern)
        updated_args = (Arg(value=SimpleString(f"'{route}'")), *other_args)
        return Call(args=updated_args, func=Name("path"))

    def build_route(self, pattern):
        """Build route from a URL pattern."""
        stripped_pattern = pattern.lstrip("^").rstrip("$")
        route = ""
        # Parse each group
        while "(?P<" in stripped_pattern:
            route_part, stripped_pattern = self.parse_next_group(stripped_pattern)
            route += route_part
        route += stripped_pattern
        self.check_route(route)
        return route

    def parse_next_group(self, left_to_parse):
        """Extract captured group info."""
        prefix, rest = left_to_parse.split("(?P<", 1)
        group, left_to_parse = rest.split(")", 1)
        group_name, group_regex = group.split(">", 1)
        try:
            converter = REGEX_TO_CONVERTER[group_regex]
        except KeyError:
            raise PatternNotSupported("No converter found")
        return prefix + f"<{converter}:{group_name}>", left_to_parse

    def check_route(self, route):
        """Check that route doesn't contain anymore regex."""
        if route != re.escape(route):
            raise PatternNotSupported("Route contains regex")



from django_codemod.constants import DJANGO_3_0, DJANGO_4_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class UGetTextTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.translation.ugettext` by `gettext`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.translation.ugettext"
    rename_to = "django.utils.translation.gettext"


class UGetTextLazyTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.translation.ugettext_lazy` by `gettext_lazy`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.translation.ugettext_lazy"
    rename_to = "django.utils.translation.gettext_lazy"


class UGetTextNoopTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.translation.ugettext_noop` by `gettext_noop`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.translation.ugettext_noop"
    rename_to = "django.utils.translation.gettext_noop"


class UNGetTextTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.translation.ungettext` by `ngettext`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.translation.ungettext"
    rename_to = "django.utils.translation.ngettext"


class UNGetTextLazyTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.translation.ungettext_lazy` by `ngettext_lazy`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.translation.ungettext_lazy"
    rename_to = "django.utils.translation.ngettext_lazy"



from libcst import BaseExpression, Call, Name
from libcst import matchers as m

from django_codemod.constants import DJANGO_2_0, DJANGO_3_0
from django_codemod.visitors.base import BaseRenameTransformer


class ContextDecoratorTransformer(BaseRenameTransformer):
    """Replace Django's `ContextDecorator` decorator by the `contextlib`'s one."""

    deprecated_in = DJANGO_2_0
    removed_in = DJANGO_3_0
    rename_from = "django.utils.decorators.ContextDecorator"
    rename_to = "contextlib.ContextDecorator"


class AvailableAttrsTransformer(BaseRenameTransformer):
    """Replace `django.utils.decorators.available_attrs` by `WRAPPER_ASSIGNMENTS`."""

    deprecated_in = DJANGO_2_0
    removed_in = DJANGO_3_0
    rename_from = "django.utils.decorators.available_attrs"
    rename_to = "functools.WRAPPER_ASSIGNMENTS"

    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        if self.is_imported_with_old_name and m.matches(
            updated_node, m.Call(func=m.Name(self.old_name))
        ):
            return Name(self.new_name)
        return super().leave_Call(original_node, updated_node)



from django_codemod.constants import DJANGO_2_0, DJANGO_3_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class AbsPathTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils._os.abspathu` by `os.path.abspath`."""

    deprecated_in = DJANGO_2_0
    removed_in = DJANGO_3_0
    rename_from = "django.utils._os.abspathu"
    rename_to = "os.path.abspath"



from django_codemod.constants import DJANGO_2_0, DJANGO_3_0, DJANGO_4_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class ForceTextTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.encoding.force_text` by `force_str`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.encoding.force_text"
    rename_to = "django.utils.encoding.force_str"


class SmartTextTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.encoding.smart_text` by `smart_str`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.encoding.smart_text"
    rename_to = "django.utils.encoding.smart_str"


class UnicodeCompatibleTransformer(BaseFuncRenameTransformer):
    """Replace Django's `python_2_unicode_compatible` by the one from `six`."""

    deprecated_in = DJANGO_2_0
    removed_in = DJANGO_3_0
    rename_from = "django.utils.encoding.python_2_unicode_compatible"
    rename_to = "six.python_2_unicode_compatible"



from parameterized import parameterized

from django_codemod.visitors import InlineHasAddPermissionsTransformer
from tests.visitors.base import BaseVisitorTest


class TestInlineHasAddPermissionsTransformer(BaseVisitorTest):

    transformer = InlineHasAddPermissionsTransformer

    def test_model_admin_base_class(self) -> None:
        """Doesn't modify if base class doesn't match."""
        before = """
            from django.contrib import admin

            class MyAdmin(admin.ModelAdmin):

                def has_add_permission(self, request):
                    if somethings:
                        return False
                    return super().has_add_permission(request)
        """
        after = """
            from django.contrib import admin

            class MyAdmin(admin.ModelAdmin):

                def has_add_permission(self, request):
                    if somethings:
                        return False
                    return super().has_add_permission(request)
        """
        self.assertCodemod(before, after)

    def test_no_base_class(self) -> None:
        """Doesn't modify if there is no base class."""
        before = """
            from django.contrib import admin

            class MyCustomStuff:

                def has_add_permission(self, request):
                    if somethings:
                        return False
                    return super().has_add_permission(request)
        """
        after = """
            from django.contrib import admin

            class MyCustomStuff:

                def has_add_permission(self, request):
                    if somethings:
                        return False
                    return super().has_add_permission(request)
        """
        self.assertCodemod(before, after)

    @parameterized.expand(
        [
            (".admin import TabularInline", "TabularInline"),
            (" import admin", "admin.TabularInline"),
            (".admin import StackedInline", "StackedInline"),
            (" import admin", "admin.StackedInline"),
        ],
    )
    def test_simple_substitution(self, import_, base_class) -> None:
        """Modification with a valid base class."""
        before = f"""
            from django.contrib{import_}

            class MyInlineInline({base_class}):

                def has_add_permission(self, request):
                    if somethings:
                        return False
                    return super().has_add_permission(request)
        """
        after = f"""
            from django.contrib{import_}

            class MyInlineInline({base_class}):

                def has_add_permission(self, request, obj = None):
                    if somethings:
                        return False
                    return super().has_add_permission(request, obj = obj)
        """
        self.assertCodemod(before, after)

    @parameterized.expand(
        [
            (".admin import TabularInline", "TabularInline"),
            (" import admin", "admin.TabularInline"),
            (".admin import StackedInline", "StackedInline"),
            (" import admin", "admin.StackedInline"),
        ],
    )
    def test_multiple_base_classes_last(self, import_, base_class) -> None:
        """Modification with multiple base classes, valid last."""
        before = f"""
            from django.contrib{import_}

            class MyInlineInline(InlineMixin, {base_class}):

                def has_add_permission(self, request):
                    if somethings:
                        return False
                    return super().has_add_permission(request)
        """
        after = f"""
            from django.contrib{import_}

            class MyInlineInline(InlineMixin, {base_class}):

                def has_add_permission(self, request, obj = None):
                    if somethings:
                        return False
                    return super().has_add_permission(request, obj = obj)
        """
        self.assertCodemod(before, after)

    @parameterized.expand(
        [
            (".admin import TabularInline", "TabularInline"),
            (" import admin", "admin.TabularInline"),
            (".admin import StackedInline", "StackedInline"),
            (" import admin", "admin.StackedInline"),
        ],
    )
    def test_multiple_base_classes_first(self, import_, base_class) -> None:
        """Modification with multiple base classes, valid first."""
        before = f"""
            from django.contrib{import_}

            class MyInlineInline({base_class}, OtherBase):

                def has_add_permission(self, request):
                    if somethings:
                        return False
                    return super().has_add_permission(request)
        """
        after = f"""
            from django.contrib{import_}

            class MyInlineInline({base_class}, OtherBase):

                def has_add_permission(self, request, obj = None):
                    if somethings:
                        return False
                    return super().has_add_permission(request, obj = obj)
        """
        self.assertCodemod(before, after)

    @parameterized.expand([("admin.TabularInline"), ("admin.StackedInline")])
    def test_context_cleared(self, base_class) -> None:
        """Test that context is cleared and doesn't leak to other classes."""
        before = f"""
            from django.contrib import admin

            class MyInlineInline({base_class}):

                def has_add_permission(self, request):
                    if somethings:
                        return False
                    return super().has_add_permission(request)


            class MyAdmin(admin.ModelAdmin):

                def has_add_permission(self, request):
                    return False
        """
        after = f"""
            from django.contrib import admin

            class MyInlineInline({base_class}):

                def has_add_permission(self, request, obj = None):
                    if somethings:
                        return False
                    return super().has_add_permission(request, obj = obj)


            class MyAdmin(admin.ModelAdmin):

                def has_add_permission(self, request):
                    return False
        """
        self.assertCodemod(before, after)



from pathlib import Path

import click
import pytest
from click.testing import CliRunner
from libcst.codemod import CodemodContext, ParallelTransformResult

from django_codemod import cli
from django_codemod.cli import (
    DEPRECATED_IN,
    REMOVED_IN,
    call_command,
    get_short_description,
)
from django_codemod.commands import BaseCodemodCommand


@pytest.fixture()
def cli_runner():
    return CliRunner()


@pytest.fixture()
def gather_files_mocked(mocker):
    """Mock return value of gather_files from libCST."""
    gather_files = mocker.patch("django_codemod.cli.gather_files")
    gather_files.return_value = ["some/file.py"]


@pytest.fixture()
def command_instance():
    """Dummy command instance to test call_command."""
    return BaseCodemodCommand([], CodemodContext())


def test_missing_argument(cli_runner):
    """Should explain missing arguments."""
    result = cli_runner.invoke(cli.djcodemod, ["run"])

    assert result.exit_code == 2
    assert "Error: Missing argument 'PATH'" in result.output


@pytest.mark.parametrize(
    "command_line",
    [
        # Missing options
        ["run", "."],
        # Too many options
        ["run", "--removed-in", "3.0", "--deprecated-in", "2.0", "."],
    ],
)
def test_invalid_options(cli_runner, command_line):
    """Should explain missing option."""
    result = cli_runner.invoke(cli.djcodemod, command_line)

    assert result.exit_code == 2
    assert (
        "Error: You must specify either '--removed-in' or "
        "'--deprecated-in' but not both." in result.output
    )


def test_help(cli_runner):
    """The --help option should be available."""
    help_result = cli_runner.invoke(cli.djcodemod, ["--help"])

    assert help_result.exit_code == 0
    assert "--help" in help_result.output
    assert "Show this message and exit." in help_result.output


def test_missing_version_parts(cli_runner):
    result = cli_runner.invoke(cli.djcodemod, ["run", "--removed-in", "3", "."])

    assert result.exit_code == 2
    assert "missing version parts." in result.output


@pytest.mark.parametrize(
    ("option", "version"),
    [
        # Removed in option
        ("--removed-in", "1.0"),
        ("--removed-in", "11.0"),
        # Deprecated in option
        ("--deprecated-in", "1.0"),
        ("--deprecated-in", "3.4"),
    ],
)
def test_non_supported_version(cli_runner, option, version):
    result = cli_runner.invoke(cli.djcodemod, ["run", option, version, "."])

    assert result.exit_code == 2
    assert f"'{version}' is not supported. Versions supported:" in result.output


def test_invalid_version(cli_runner):
    result = cli_runner.invoke(
        cli.djcodemod, ["run", "--removed-in", "not.a.version", "."]
    )

    assert result.exit_code == 2
    assert "'not.a.version' is not a valid version" in result.output


@pytest.mark.parametrize(
    ("option", "version"),
    [
        # Removed in option
        ("--removed-in", "3.0"),
        ("--deprecated-in", "2.0"),
    ],
)
@pytest.mark.parametrize("path", [".", "my_app"])
def test_basic_arguments(cli_runner, option, version, path):
    with cli_runner.isolated_filesystem() as tempdir:
        dir = Path(tempdir) / "my_app"
        dir.mkdir()
        for file_name in ["hello1.py", "hello2.py", "hello3.py"]:
            py_file = dir / file_name
            with py_file.open("w") as f:
                f.write('print("Hello World!")')

        result = cli_runner.invoke(cli.djcodemod, ["run", option, version, path])

    assert result.exit_code == 0, result
    assert "Finished codemodding 3 files!" in result.output


@pytest.mark.usefixtures("gather_files_mocked")
def test_call_command_success(command_instance, mocker):
    executor = mocker.patch(
        "django_codemod.cli.parallel_exec_transform_with_prettyprint"
    )
    executor.return_value = ParallelTransformResult(
        successes=1, failures=0, warnings=0, skips=0
    )

    result = call_command(command_instance, ".")

    assert result is None


@pytest.mark.usefixtures("gather_files_mocked")
def test_call_command_failure(command_instance, mocker):
    executor = mocker.patch(
        "django_codemod.cli.parallel_exec_transform_with_prettyprint"
    )
    executor.return_value = ParallelTransformResult(
        successes=0, failures=1, warnings=0, skips=0
    )

    with pytest.raises(click.exceptions.Exit):
        call_command(command_instance, ".")


@pytest.mark.usefixtures("gather_files_mocked")
def test_call_command_interrupted(command_instance, mocker):
    executor = mocker.patch(
        "django_codemod.cli.parallel_exec_transform_with_prettyprint",
        side_effect=KeyboardInterrupt(),
    )
    executor.return_value = ParallelTransformResult(
        successes=1, failures=0, warnings=0, skips=0
    )

    with pytest.raises(click.Abort):
        call_command(command_instance, ".")


def _mapping_repr(mapping):
    """Helper to return class names in the dict values."""
    return {
        version: [klass.__name__ for klass in classes_list]
        for version, classes_list in mapping.items()
    }


def test_deprecated_in_mapping():
    """Transformers found by the ``DEPRECATED_IN`` mapping."""
    assert _mapping_repr(DEPRECATED_IN) == {
        (3, 0): [
            "ForceTextTransformer",
            "HttpUrlQuotePlusTransformer",
            "HttpUrlQuoteTransformer",
            "HttpUrlUnQuotePlusTransformer",
            "HttpUrlUnQuoteTransformer",
            "IsSafeUrlTransformer",
            "SmartTextTransformer",
            "UGetTextLazyTransformer",
            "UGetTextNoopTransformer",
            "UGetTextTransformer",
            "UNGetTextLazyTransformer",
            "UNGetTextTransformer",
            "URLTransformer",
            "UnescapeEntitiesTransformer",
        ],
        (2, 1): ["InlineHasAddPermissionsTransformer"],
        (2, 0): [
            "AbsPathTransformer",
            "AvailableAttrsTransformer",
            "ContextDecoratorTransformer",
            "LRUCacheTransformer",
            "RenderToResponseTransformer",
            "UnicodeCompatibleTransformer",
        ],
        (1, 11): ["ModelsPermalinkTransformer"],
        (1, 10): ["URLResolversTransformer"],
        (1, 9): ["OnDeleteTransformer"],
    }


def test_removed_in_mapping():
    """Transformers found by the ``REMOVED_IN`` mapping."""
    assert _mapping_repr(REMOVED_IN) == {
        (4, 0): [
            "ForceTextTransformer",
            "HttpUrlQuotePlusTransformer",
            "HttpUrlQuoteTransformer",
            "HttpUrlUnQuotePlusTransformer",
            "HttpUrlUnQuoteTransformer",
            "IsSafeUrlTransformer",
            "SmartTextTransformer",
            "UGetTextLazyTransformer",
            "UGetTextNoopTransformer",
            "UGetTextTransformer",
            "UNGetTextLazyTransformer",
            "UNGetTextTransformer",
            "URLTransformer",
            "UnescapeEntitiesTransformer",
        ],
        (3, 0): [
            "AbsPathTransformer",
            "AvailableAttrsTransformer",
            "ContextDecoratorTransformer",
            "InlineHasAddPermissionsTransformer",
            "LRUCacheTransformer",
            "RenderToResponseTransformer",
            "UnicodeCompatibleTransformer",
        ],
        (2, 1): ["ModelsPermalinkTransformer"],
        (2, 0): ["OnDeleteTransformer", "URLResolversTransformer"],
    }


def test_list_command(cli_runner):
    result = cli_runner.invoke(cli.djcodemod, ["list"])

    assert result.exit_code == 0, result
    assert "Codemodder" in result.output
    assert "Deprecated in" in result.output
    assert "Removed in" in result.output
    assert "Description" in result.output


class NoDocString:
    pass


class SingleLine:
    """This is the description."""


class MultiLine1:
    """This is the title.

    This is more details."""


class MultiLine2:
    """
    Another title.

    And even more extra stuff.
    """


@pytest.mark.parametrize(
    ("klass", "expected_result"),
    [
        (NoDocString, ""),
        (SingleLine, "This is the description."),
        (MultiLine1, "This is the title."),
        (MultiLine2, "Another title."),
    ],
)
def test_get_short_description(klass, expected_result):
    assert get_short_description(klass) == expected_result



import inspect
from collections import defaultdict
from operator import attrgetter
from typing import Callable, Dict, Generator, List, Tuple, Type

import click
from libcst.codemod import (
    CodemodContext,
    gather_files,
    parallel_exec_transform_with_prettyprint,
)
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table

from django_codemod import visitors
from django_codemod.commands import BaseCodemodCommand
from django_codemod.visitors.base import BaseDjCodemodTransformer


def index_codemoders(version_getter: Callable) -> Dict[Tuple[int, int], List]:
    """
    Index codemodders by version.

    Build a map of Django version to list of codemodders.
    """
    codemodders_index = defaultdict(list)
    for obj in iter_codemodders():
        django_version = version_getter(obj)
        codemodders_index[django_version].append(obj)
    return dict(codemodders_index)


def iter_codemodders() -> Generator[BaseDjCodemodTransformer, None, None]:
    """Iterator of all the codemodders classes."""
    for object_name in dir(visitors):
        try:
            obj = getattr(visitors, object_name)
            if not issubclass(obj, BaseDjCodemodTransformer) or inspect.isabstract(obj):
                continue
            # Looks like this one is good to go
            yield obj
        except TypeError:
            continue


DEPRECATED_IN = index_codemoders(version_getter=attrgetter("deprecated_in"))
REMOVED_IN = index_codemoders(version_getter=attrgetter("removed_in"))


class VersionParamType(click.ParamType):
    """A type of parameter to parse version as arguments."""

    name = "version"
    example = (
        "Should include the major & minor digits of the Django version"
        " e.g. '2.2' or '2.2.10'"
    )

    def __init__(self, version_index) -> None:
        self.valid_versions = version_index.keys()

    def convert(self, value, param, ctx):
        """Parse version to keep only major an minor digits."""
        try:
            return self._parse_unsafe(value, param, ctx)
        except ValueError:
            self.fail(f"{value!r} is not a valid version. {self.example}", param, ctx)

    def _parse_unsafe(self, value, param, ctx):
        """Parse version and validate it's a supported one."""
        parsed_version = self._split_digits(value, param, ctx)
        if parsed_version not in self.valid_versions:
            supported_versions = ", ".join(
                ".".join(str(version_part) for version_part in version_tuple)
                for version_tuple in self.valid_versions
            )
            self.fail(
                f"{value!r} is not supported. "
                f"Versions supported: {supported_versions}",
                param,
                ctx,
            )
        return parsed_version

    def _split_digits(self, value, param, ctx):
        """Split version into 2-tuple of digits, ignoring patch digit."""
        values_parts = tuple(int(v) for v in value.split("."))
        if len(values_parts) < 2:
            self.fail(
                f"{value!r} missing version parts. {self.example}",
                param,
                ctx,
            )
        major, minor, *patches = values_parts
        return (major, minor)


@click.group()
def djcodemod():
    """CLI entry point."""


@djcodemod.command()
@click.argument("path")
@click.option(
    "--removed-in",
    "removed_in",
    help="The version of Django where feature are removed.",
    type=VersionParamType(REMOVED_IN),
)
@click.option(
    "--deprecated-in",
    "deprecated_in",
    help="The version of Django where deprecations started.",
    type=VersionParamType(DEPRECATED_IN),
)
def run(removed_in: Tuple[int, int], deprecated_in: Tuple[int, int], path: str) -> None:
    """
    Automatically fixes deprecations removed Django deprecations.

    This command takes the path to target as argument and a version of
    Django to select code modifications to apply.
    """
    if not any((removed_in, deprecated_in)) or all((removed_in, deprecated_in)):
        raise click.UsageError(
            "You must specify either '--removed-in' or '--deprecated-in' but not both."
        )
    if removed_in:
        codemodders_list = REMOVED_IN[removed_in]
    else:
        codemodders_list = DEPRECATED_IN[deprecated_in]
    command_instance = BaseCodemodCommand(codemodders_list, CodemodContext())
    call_command(command_instance, path)


def call_command(command_instance: BaseCodemodCommand, path: str):
    """Call libCST with our customized command."""
    files = gather_files([path])
    try:
        # Super simplified call
        result = parallel_exec_transform_with_prettyprint(
            command_instance,
            files,
            # Number of jobs to use when processing files. Defaults to number of cores
            jobs=None,
        )
    except KeyboardInterrupt:
        raise click.Abort("Interrupted!")

    # fancy summary a-la libCST
    total = result.successes + result.skips + result.failures
    click.echo(f"Finished codemodding {total} files!")
    click.echo(f" - Transformed {result.successes} files successfully.")
    click.echo(f" - Skipped {result.skips} files.")
    click.echo(f" - Failed to codemod {result.failures} files.")
    click.echo(f" - {result.warnings} warnings were generated.")
    if result.failures > 0:
        raise click.exceptions.Exit(1)


@djcodemod.command()
def list() -> None:
    """Print all available codemodders as a table."""
    console = Console()
    table = Table(show_header=True, header_style="bold")
    # Headers
    table.add_column("Codemodder")
    table.add_column("Deprecated in", justify="right")
    table.add_column("Removed in", justify="right")
    table.add_column("Description")
    # Content
    prev_version = None
    for name, deprecated_in, removed_in, description in generate_rows():
        if prev_version and prev_version != (deprecated_in, removed_in):
            table.add_row()
        table.add_row(name, deprecated_in, removed_in, Markdown(description))
        prev_version = (deprecated_in, removed_in)
    # Print it out
    console.print(table)


def generate_rows() -> Generator[Tuple[str, str, str, str], None, None]:
    """Build up the rows for the table of codemodders."""
    codemodders_list = sorted(
        iter_codemodders(), key=lambda obj: (obj.deprecated_in, obj.removed_in)
    )
    for codemodder in codemodders_list:
        yield (
            codemodder.__name__,
            version_str(codemodder.deprecated_in),
            version_str(codemodder.removed_in),
            get_short_description(codemodder),
        )


def get_short_description(codemodder: Type) -> str:
    """Get a one line description of the codemodder from its docstring."""
    if codemodder.__doc__ is None:
        return ""
    for line in codemodder.__doc__.split("\n"):
        description = line.strip()
        if description:
            return description


def version_str(version_parts: Tuple[int, int]) -> str:
    """Format the version tuple as string."""
    return ".".join(str(d) for d in version_parts)



from django_codemod.constants import DJANGO_3_0, DJANGO_4_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class HttpUrlQuoteTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.urlquote` by `urllib.parse.quote`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.urlquote"
    rename_to = "urllib.parse.quote"


class HttpUrlQuotePlusTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.urlquote_plus` by `urllib.parse.quote_plus`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.urlquote_plus"
    rename_to = "urllib.parse.quote_plus"


class HttpUrlUnQuoteTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.urlunquote` by `urllib.parse.unquote`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.urlunquote"
    rename_to = "urllib.parse.unquote"


class HttpUrlUnQuotePlusTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.http.urlunquote_plus` by `urllib.parse.unquote_plus`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.urlunquote_plus"
    rename_to = "urllib.parse.unquote_plus"


class IsSafeUrlTransformer(BaseFuncRenameTransformer):
    """Rename `django.utils.http.is_safe_url` to `url_has_allowed_host_and_scheme`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.http.is_safe_url"
    rename_to = "django.utils.http.url_has_allowed_host_and_scheme"



from django_codemod.visitors import UnescapeEntitiesTransformer
from tests.visitors.base import BaseVisitorTest


class TestUnescapeEntitiesTransformer(BaseVisitorTest):

    transformer = UnescapeEntitiesTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.text import unescape_entities

            result = unescape_entities(content)
        """
        after = """
            from html import unescape

            result = unescape(content)
        """
        self.assertCodemod(before, after)



from django_codemod.visitors import URLTransformer
from tests.visitors.base import BaseVisitorTest


class TestURLTransformer(BaseVisitorTest):

    transformer = URLTransformer

    def test_noop(self) -> None:
        """Test when nothing should change."""
        before = """
            from django.urls import include, re_path

            urlpatterns = [
                re_path(r'^index/$', views.index, name='index'),
                re_path(r'^weblog/', include('blog.urls')),
            ]
        """
        after = """
            from django.urls import include, re_path

            urlpatterns = [
                re_path(r'^index/$', views.index, name='index'),
                re_path(r'^weblog/', include('blog.urls')),
            ]
        """

        self.assertCodemod(before, after)

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.conf.urls import url
            from django.urls import include

            urlpatterns = [
                url(r'^$', views.index, name='index'),
                url(r'^about/$', views.about, name='about'),
                url(r'^post/(?P<slug>[w-]+)/$', views.post, name='post'),
                url(r'^weblog/', include('blog.urls')),
            ]
        """
        after = """
            from django.urls import path, re_path, include

            urlpatterns = [
                path('', views.index, name='index'),
                path('about/', views.about, name='about'),
                re_path(r'^post/(?P<slug>[w-]+)/$', views.post, name='post'),
                path('weblog/', include('blog.urls')),
            ]
        """
        self.assertCodemod(before, after)

    def test_starting_caret(self) -> None:
        """Patterns not starting with '^' are migrated to re_path."""
        before = """
            from django.conf.urls import url

            urlpatterns = [
                url(r'about/$', views.about, name='about'),
            ]
        """
        after = """
            from django.urls import re_path

            urlpatterns = [
                re_path(r'about/$', views.about, name='about'),
            ]
        """
        self.assertCodemod(before, after)

    def test_all_path(self) -> None:
        """Check when all can be replaced with path."""
        before = """
            from django.conf.urls import url
            from django.urls import include

            urlpatterns = [
                url(r'^$', views.index, name='index'),
                url(r'^about/$', views.about, name='about'),
            ]
        """
        after = """
            from django.urls import path, include

            urlpatterns = [
                path('', views.index, name='index'),
                path('about/', views.about, name='about'),
            ]
        """
        self.assertCodemod(before, after)

    def test_translated_pattern(self):
        """URL patterns can also be marked translatable."""
        before = """
            from django.urls import include
            from django.conf.urls import url
            from django.utils.translation import gettext_lazy as _

            urlpatterns = [
                url(_(r'^about/$'), views.about, name='about'),
                url(_(r'^post/(?P<slug>[w-]+)/$'), views.post, name='post'),
                url(_(r'^weblog/'), include('blog.urls')),
            ]
        """
        after = """
            from django.urls import re_path, include
            from django.utils.translation import gettext_lazy as _

            urlpatterns = [
                re_path(_(r'^about/$'), views.about, name='about'),
                re_path(_(r'^post/(?P<slug>[w-]+)/$'), views.post, name='post'),
                re_path(_(r'^weblog/'), include('blog.urls')),
            ]
        """
        self.assertCodemod(before, after)

    def test_grouped_path(self) -> None:
        """Check replacing pattern with groups."""
        uuid_re = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
        before = fr"""
            from django.conf.urls import url

            urlpatterns = [
                url(r'^page/(?P<number>[0-9]+)/$', views.page, name='page'),
                url(r'^page/(?P<number>\d+)/$', views.page, name='page'),
                url(r'^post/(?P<slug>[-a-zA-Z0-9_]+)/$', views.post, name='post'),
                url(r'^post/(?P<slug>[\w-]+)/$', views.post, name='post'),
                url(r'^about/(?P<name>[^/]+)/$', views.about, name='about'),
                url(r'^uuid/(?P<uuid>{uuid_re})/$', by_uuid),
                url(r'^(?P<path>.+)/$', views.default, name='default'),
            ]
        """
        after = """
            from django.urls import path

            urlpatterns = [
                path('page/<int:number>/', views.page, name='page'),
                path('page/<int:number>/', views.page, name='page'),
                path('post/<slug:slug>/', views.post, name='post'),
                path('post/<slug:slug>/', views.post, name='post'),
                path('about/<str:name>/', views.about, name='about'),
                path('uuid/<uuid:uuid>/', by_uuid),
                path('<path:path>/', views.default, name='default'),
            ]
        """
        self.assertCodemod(before, after)

    def test_non_captured_group(self) -> None:
        """Non-capturing regex should be moved to re_path."""
        before = """
            from django.conf.urls import url

            urlpatterns = [
                url(r'^page/[0-9]+/$', views.page, name='page'),
            ]
        """
        after = """
            from django.urls import re_path

            urlpatterns = [
                re_path(r'^page/[0-9]+/$', views.page, name='page'),
            ]
        """
        self.assertCodemod(before, after)



from django_codemod.visitors import (
    UGetTextLazyTransformer,
    UGetTextNoopTransformer,
    UGetTextTransformer,
    UNGetTextLazyTransformer,
    UNGetTextTransformer,
)
from tests.visitors.base import BaseVisitorTest


class TestUGetTextTransformer(BaseVisitorTest):

    transformer = UGetTextTransformer

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.utils.translation import ugettext

            result = ugettext(content)
        """
        after = """
            from django.utils.translation import gettext

            result = gettext(content)
        """
        self.assertCodemod(before, after)

    def test_import_with_alias(self) -> None:
        """Check case with a common alias."""
        before = """
            from django.utils.translation import ugettext as _

            result = _(content)
        """
        after = """
            from django.utils.translation import gettext as _

            result = _(content)
        """
        self.assertCodemod(before, after)

    def test_already_imported_substitution(self) -> None:
        """Test case where gettext is already in the imports."""
        before = """
            from django.utils.translation import ugettext, gettext

            result = ugettext(content)
        """
        after = """
            from django.utils.translation import gettext

            result = gettext(content)
        """
        self.assertCodemod(before, after)


class TestUGetTextLazyTransformer(BaseVisitorTest):

    transformer = UGetTextLazyTransformer

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.utils.translation import ugettext_lazy

            result = ugettext_lazy(content)
        """
        after = """
            from django.utils.translation import gettext_lazy

            result = gettext_lazy(content)
        """
        self.assertCodemod(before, after)


class TestUGetTextNoopTransformer(BaseVisitorTest):

    transformer = UGetTextNoopTransformer

    def test_noop(self) -> None:
        """Test when nothing should change."""
        before = """
            from django import conf
            from django.utils import translation

            foo = gettext_noop("bar")
        """
        after = """
            from django import conf
            from django.utils import translation

            foo = gettext_noop("bar")
        """

        self.assertCodemod(before, after)

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.utils.translation import ugettext_noop

            result = ugettext_noop(content)
        """
        after = """
            from django.utils.translation import gettext_noop

            result = gettext_noop(content)
        """
        self.assertCodemod(before, after)


class TestUNGetTextTransformer(BaseVisitorTest):

    transformer = UNGetTextTransformer

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.utils.translation import ungettext

            result = ungettext(content, plural_content, count)
        """
        after = """
            from django.utils.translation import ngettext

            result = ngettext(content, plural_content, count)
        """
        self.assertCodemod(before, after)


class TestUNGetTextLazyTransformer(BaseVisitorTest):

    transformer = UNGetTextLazyTransformer

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.utils.translation import ungettext_lazy

            result = ungettext_lazy(content, plural_content, count)
        """
        after = """
            from django.utils.translation import ngettext_lazy

            result = ngettext_lazy(content, plural_content, count)
        """
        self.assertCodemod(before, after)

    def test_import_as_alias(self) -> None:
        """Check with a common import alias."""
        before = """
            from django.utils.translation import ungettext_lazy as _

            result = _(content, plural_content, count)
        """
        after = """
            from django.utils.translation import ngettext_lazy as _

            result = _(content, plural_content, count)
        """
        self.assertCodemod(before, after)



from django_codemod.constants import DJANGO_3_0, DJANGO_4_0
from django_codemod.visitors.base import BaseFuncRenameTransformer


class UnescapeEntitiesTransformer(BaseFuncRenameTransformer):
    """Replace `django.utils.text.unescape_entities` by `html.unescape`."""

    deprecated_in = DJANGO_3_0
    removed_in = DJANGO_4_0
    rename_from = "django.utils.text.unescape_entities"
    rename_to = "html.unescape"



from abc import ABC
from typing import List, Type

import libcst as cst
from libcst.codemod import (
    CodemodContext,
    ContextAwareTransformer,
    VisitorBasedCodemodCommand,
)


class BaseCodemodCommand(VisitorBasedCodemodCommand, ABC):
    """Base class for our commands."""

    transformers: List[Type[ContextAwareTransformer]]

    def __init__(self, transformers, context: CodemodContext) -> None:
        self.transformers = transformers
        super().__init__(context)

    def transform_module_impl(self, tree: cst.Module) -> cst.Module:
        for transform in self.transformers:
            inst = transform(self.context)
            tree = inst.transform_module(tree)
        return tree



from django_codemod.visitors import (
    ForceTextTransformer,
    SmartTextTransformer,
    UnicodeCompatibleTransformer,
)
from tests.visitors.base import BaseVisitorTest


class TestForceTextTransformer(BaseVisitorTest):

    transformer = ForceTextTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.utils.encoding import force_text

            result = force_text(content)
        """
        after = """
            from django.utils.encoding import force_str

            result = force_str(content)
        """
        self.assertCodemod(before, after)


class TestSmartTextTransformer(BaseVisitorTest):

    transformer = SmartTextTransformer

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.utils.encoding import smart_text

            result = smart_text(content)
        """
        after = """
            from django.utils.encoding import smart_str

            result = smart_str(content)
        """
        self.assertCodemod(before, after)


class TestUnicodeCompatibleTransformer(BaseVisitorTest):

    transformer = UnicodeCompatibleTransformer

    def test_simple_substitution(self) -> None:
        """Check simple use case."""
        before = """
            from django.utils.encoding import python_2_unicode_compatible

            @python_2_unicode_compatible
            class MyModel:
                pass
        """
        after = """
            from six import python_2_unicode_compatible

            @python_2_unicode_compatible
            class MyModel:
                pass
        """
        self.assertCodemod(before, after)



import pytest
from libcst import matchers as m

from django_codemod.visitors.base import (
    BaseFuncRenameTransformer,
    BaseModuleRenameTransformer,
    module_matcher,
)

from .base import BaseVisitorTest


@pytest.mark.parametrize(
    ("parts", "expected_matcher"),
    [
        (["django"], m.Attribute(attr=m.Name("django"))),
        (
            ["django", "contrib"],
            m.Attribute(value=m.Name("django"), attr=m.Name("contrib")),
        ),
        (
            ["django", "contrib", "admin"],
            m.Attribute(
                value=m.Attribute(value=m.Name("django"), attr=m.Name("contrib")),
                attr=m.Name("admin"),
            ),
        ),
    ],
)
def test_module_matcher(parts, expected_matcher):
    matcher = module_matcher(parts)

    # equality comparision doesn't work with matcher:
    # compare their representation seems to work
    assert repr(matcher) == repr(expected_matcher)


class SameModuleFuncRenameTransformer(BaseFuncRenameTransformer):
    """Simple transformer renaming function from same module."""

    rename_from = "django.dummy.module.func"
    rename_to = "django.dummy.module.better_func"


class TestFuncRenameTransformer(BaseVisitorTest):

    transformer = SameModuleFuncRenameTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.dummy.module import func

            result = func()
        """
        after = """
            from django.dummy.module import better_func

            result = better_func()
        """
        self.assertCodemod(before, after)

    def test_already_imported(self) -> None:
        """Function to modify is already imported with an alias."""
        before = """
            from django.dummy.module import func, better_func

            result = func(content)
        """
        after = """
            from django.dummy.module import better_func

            result = better_func(content)
        """
        self.assertCodemod(before, after)

    def test_import_with_alias(self) -> None:
        """Function to modify is imported with an alias."""
        before = """
            from django.dummy.module import func as aliased_func

            result = aliased_func()
        """
        after = """
            from django.dummy.module import better_func as aliased_func

            result = aliased_func()
        """
        self.assertCodemod(before, after)

    def test_same_name_function(self) -> None:
        """Should not be fooled by a function bearing the same name."""
        before = """
            from utils.helpers import func

            result = func()
        """
        after = """
            from utils.helpers import func

            result = func()
        """
        self.assertCodemod(before, after)

    def test_same_name_with_alias_import_function(self) -> None:
        """Imported with alias and other function with the same name."""
        before = """
            from django.dummy.module import func as aliased_func
            from utils.helpers import func

            result = func()
            aliased_func()
        """
        after = """
            from utils.helpers import func
            from django.dummy.module import better_func as aliased_func

            result = func()
            aliased_func()
        """
        self.assertCodemod(before, after)

    def test_extra_trailing_comma_when_last(self) -> None:
        """Extra trailing comma when removed import is the last one."""
        before = """
            from django.dummy.module import better_func, func

            result = func(content)
        """
        after = """
            from django.dummy.module import better_func

            result = better_func(content)
        """
        self.assertCodemod(before, after)

    def test_parse_call_no_value(self) -> None:
        """Bug with function call without name."""
        before = """
            factory()()
        """
        after = """
            factory()()
        """
        self.assertCodemod(before, after)

    def test_lambda_no_value(self) -> None:
        """Bug with lambda call without name."""
        before = """
            (lambda x: x)(something)
        """
        after = """
            (lambda x: x)(something)
        """
        self.assertCodemod(before, after)


class OtherModuleFuncRenameTransformer(BaseFuncRenameTransformer):
    """Transformer with different module."""

    rename_from = "django.dummy.module.func"
    rename_to = "django.better.dummy.better_func"


class TestOtherModuleFuncRenameTransformer(BaseVisitorTest):

    transformer = OtherModuleFuncRenameTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.dummy.module import func

            result = func()
        """
        after = """
            from django.better.dummy import better_func

            result = better_func()
        """
        self.assertCodemod(before, after)

    def test_already_imported(self) -> None:
        before = """
            from django.dummy.module import func
            from django.better.dummy import better_func

            result = func(content)
        """
        after = """
            from django.better.dummy import better_func

            result = better_func(content)
        """
        self.assertCodemod(before, after)

    def test_import_with_alias(self) -> None:
        before = """
            from django.dummy.module import func as aliased_func

            result = aliased_func()
        """
        after = """
            from django.better.dummy import better_func as aliased_func

            result = aliased_func()
        """
        self.assertCodemod(before, after)


class OtherModuleRenameTransformer(BaseModuleRenameTransformer):
    """Simple transformer renaming function from same module."""

    rename_from = "django.dummy.module"
    rename_to = "django.dummy.other_module"


class TestModuleRenameTransformer(BaseVisitorTest):

    transformer = OtherModuleRenameTransformer

    def test_simple_substitution(self) -> None:
        before = """
            from django.dummy.module import func

            result = func()
        """
        after = """
            from django.dummy.other_module import func

            result = func()
        """
        self.assertCodemod(before, after)
